# Copyright 2019 Open Source Robotics Foundation, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from ros2interface.verb import VerbExtension
from rosidl_runtime_py import get_action_interfaces
from rosidl_runtime_py import get_interface_packages
from rosidl_runtime_py import get_message_interfaces
from rosidl_runtime_py import get_service_interfaces


def print_packages(packs):
    for name in sorted(packs):
        print((name))


class PackagesVerb(VerbExtension):
    """Output a list of packages that provide interfaces."""

    def add_arguments(self, parser, cli_name):
        parser.add_argument('-m', '--only-msgs', action='count',
                            help='Only list packages that generate messages')

        parser.add_argument('-s', '--only-srvs', action='count',
                            help='Only list packages that generate services')

        parser.add_argument('-a', '--only-actions', action='count',
                            help='Only list packages that generate actions')

    def main(self, *, args):
        if args.only_msgs:
            print_packages(sorted(get_message_interfaces()))
        elif args.only_srvs:
            print_packages(sorted(get_service_interfaces()))
        elif args.only_actions:
            print_packages(sorted(get_action_interfaces()))
        else:
            print_packages(sorted(get_interface_packages()))



# Copyright 2019 Canonical Ldt.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from ros2interface.api import interface_to_yaml
from ros2interface.api import type_completer
from ros2interface.verb import VerbExtension


class ProtoVerb(VerbExtension):
    """Output an interface prototype."""

    def add_arguments(self, parser, cli_name):
        arg = parser.add_argument(
                'type',
                help="Show an interface definition (e.g. 'example_interfaces/msg/String')")
        arg.completer = type_completer
        parser.add_argument(
            '--no-quotes', action='store_true',
            help='if true output has no outer quotes.')

    def main(self, *, args):
        yaml = interface_to_yaml(args.type)

        if args.no_quotes is True:
            print(yaml)
        else:
            print('"' + yaml + '"')



# Copyright 2019 Open Source Robotics Foundation, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from ros2interface.api import package_name_completer
from ros2interface.verb import VerbExtension
from rosidl_runtime_py import get_interfaces


class PackageVerb(VerbExtension):
    """Output a list of available interface types within one package."""

    def add_arguments(self, parser, cli_name):
        arg = parser.add_argument(
            'package_name',
            help="Name of the ROS package (e.g. 'example_interfaces')")
        arg.completer = package_name_completer

    def main(self, *, args):
        try:
            interfaces = get_interfaces([args.package_name])
        except LookupError as e:
            return str(e)
        for package_name in sorted(interfaces):
            for interface_name in interfaces[package_name]:
                print(f'{package_name}/{interface_name}')



# Copyright 2019 Open Source Robotics Foundation, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import argparse
import sys
import typing

from ros2interface.api import type_completer
from ros2interface.verb import VerbExtension
from rosidl_adapter.parser import \
    ACTION_REQUEST_RESPONSE_SEPARATOR, \
    Constant, \
    Field, \
    MessageSpecification, \
    parse_message_string, \
    SERVICE_REQUEST_RESPONSE_SEPARATOR
from rosidl_runtime_py import get_interface_path


class InterfaceTextLine:
    """A convenience class for a single text line in an interface file."""

    def __init__(
        self,
        pkg_name: str,
        msg_name: str,
        line_text: str,
    ):
        if line_text in (SERVICE_REQUEST_RESPONSE_SEPARATOR, ACTION_REQUEST_RESPONSE_SEPARATOR):
            msg_spec = None
        else:
            msg_spec = parse_message_string(
                pkg_name=pkg_name,
                msg_name=msg_name,
                message_string=line_text,
            )
            if len(msg_spec.fields) > 1:
                raise ValueError("'line_text' must be only one line")
        self._msg_spec: typing.Optional[MessageSpecification] = msg_spec
        self._raw_line_text = line_text

    def __str__(self) -> str:
        return self._raw_line_text

    def is_comment(self) -> bool:
        return self._msg_spec and self._msg_spec.annotations['comment']

    def is_trailing_comment(self) -> bool:
        return (
            self._is_field_trailing_comment()
            or self._is_constant_trailing_comment()
        )

    def _is_field_trailing_comment(self) -> bool:
        return self._field and self._field.annotations['comment']

    def _is_constant_trailing_comment(self) -> bool:
        return self._constant and self._constant.annotations['comment']

    @property
    def nested_type(self) -> typing.Optional[str]:
        if self._field and self._is_nested():
            interface_type: str = str(self._field.type)
            if self._field.type.is_array:
                interface_type = interface_type[:interface_type.find('[')]
            return interface_type.replace('/', '/msg/')

    @property
    def trailing_comment(self) -> typing.Optional[str]:
        if self._is_field_trailing_comment():
            return self._field.annotations['comment'][0]
        elif self._is_constant_trailing_comment():
            return self._constant.annotations['comment'][0]
        else:
            return None

    @property
    def _field(self) -> typing.Optional[Field]:
        if self._msg_spec and self._msg_spec.fields:
            return self._msg_spec.fields[0]

    @property
    def _constant(self) -> typing.Optional[Constant]:
        if self._msg_spec and self._msg_spec.constants:
            return self._msg_spec.constants[0]

    def _is_nested(self) -> bool:
        if self._msg_spec and self._msg_spec.fields:
            return '/' in str(self._field.type)
        else:
            return False


def _get_interface_lines(interface_identifier: str) -> typing.Iterable[InterfaceTextLine]:
    parts: typing.List[str] = interface_identifier.split('/')
    if len(parts) != 3:
        raise ValueError(
            f"Invalid name '{interface_identifier}'. Expected three parts separated by '/'"
        )
    pkg_name, _, msg_name = parts

    file_path = get_interface_path(interface_identifier)
    with open(file_path) as file_handler:
        for line in file_handler:
            yield InterfaceTextLine(
                pkg_name=pkg_name,
                msg_name=msg_name,
                line_text=line.rstrip(),
            )


def _print_interface_line(
    line: InterfaceTextLine,
    is_show_comments: bool,
    indent_level: int
):
    text = str(line)
    if not is_show_comments:
        if not text or line.is_comment():
            return
        elif line.is_trailing_comment():
            comment_start_idx = text.find(line.trailing_comment)
            text = text[:comment_start_idx - 1].strip()
    if text:
        indent_string = indent_level * '\t'
        print(f'{indent_string}{text}')
    else:
        print()


def _show_interface(
    interface_identifier: str,
    is_show_comments: bool = False,
    is_show_nested_comments: bool = False,
    indent_level: int = 0
):
    for line in _get_interface_lines(interface_identifier):

        _print_interface_line(
            line, is_show_comments=is_show_comments, indent_level=indent_level)

        if line.nested_type:
            _show_interface(
                line.nested_type,
                is_show_comments=is_show_nested_comments,
                is_show_nested_comments=is_show_nested_comments,
                indent_level=indent_level+1,
            )


class ReadStdinPipe(argparse.Action):
    """Get argument from stdin pipe."""

    def __call__(self, parser, namespace, values, option_string=None):
        if values == '-':
            if sys.stdin.isatty():
                parser.error('expected stdin pipe')
            values = sys.stdin.readline().strip()
        if not values:
            parser.error('the passed value is empty')
        setattr(namespace, self.dest, values)


class ShowVerb(VerbExtension):
    """Output the interface definition."""

    def add_arguments(self, parser, cli_name):
        comment_group = parser.add_mutually_exclusive_group()
        comment_group.add_argument(
            '--all-comments',
            default=False, action='store_true',
            help='Show all comments, including for nested interface definitions'
        )
        comment_group.add_argument(
            '--no-comments',
            default=False, action='store_true',
            help='Show no comments or whitespace'
        )

        arg = parser.add_argument(
            'type',
            action=ReadStdinPipe,
            help="Show an interface definition (e.g. 'example_interfaces/msg/String'). "
                 "Passing '-' reads the argument from stdin (e.g. "
                 "'ros2 topic type /chatter | ros2 interface show -').")
        arg.completer = type_completer

    def main(self, *, args):
        try:
            _show_interface(
                args.type,
                is_show_comments=not args.no_comments,
                is_show_nested_comments=args.all_comments,
            )
        except (ValueError, LookupError) as e:
            return str(e)



# Copyright 2019 Open Source Robotics Foundation, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from ros2interface.verb import VerbExtension
from rosidl_runtime_py import get_action_interfaces
from rosidl_runtime_py import get_message_interfaces
from rosidl_runtime_py import get_service_interfaces


def print_messages():
    print('Messages:')
    message_interfaces = get_message_interfaces()
    for package_name in sorted(message_interfaces):
        for message_name in sorted(message_interfaces[package_name]):
            print(f'    {package_name}/{message_name}')


def print_services():
    print('Services:')
    service_interfaces = get_service_interfaces()
    for package_name in sorted(service_interfaces):
        for service_name in sorted(service_interfaces[package_name]):
            print(f'    {package_name}/{service_name}')


def print_actions():
    print('Actions:')
    action_interfaces = get_action_interfaces()
    for package_name in sorted(action_interfaces):
        for action_name in sorted(action_interfaces[package_name]):
            print(f'    {package_name}/{action_name}')


class ListVerb(VerbExtension):
    """List all interface types available."""

    def add_arguments(self, parser, cli_name):
        parser.add_argument(
            '-m', '--only-msgs', action='store_true',
            help='Print out only the message types')

        parser.add_argument(
            '-s', '--only-srvs', action='store_true',
            help='Print out only the service types')

        parser.add_argument(
            '-a', '--only-actions', action='store_true',
            help='Print out only the action types')

    def main(self, *, args):
        if args.only_msgs or args.only_srvs or args.only_actions:
            if args.only_msgs:
                print_messages()
            if args.only_srvs:
                print_services()
            if args.only_actions:
                print_actions()
            return
        print_messages()
        print_services()
        print_actions()



# Copyright 2019 Open Source Robotics Foundation, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from rosidl_runtime_py import get_action_interfaces
from rosidl_runtime_py import get_interface_packages
from rosidl_runtime_py import get_message_interfaces
from rosidl_runtime_py import get_service_interfaces
from rosidl_runtime_py import message_to_yaml
from rosidl_runtime_py import utilities


def package_name_completer(**kwargs):
    """Callable returning a list of packages containing messages, services, and action."""
    return get_interface_packages()


def type_completer(**kwargs):
    """Callable returning a list of message, service, and action types."""
    types = []
    for package_name, service_names in get_service_interfaces().items():
        for service_name in service_names:
            types.append(f'{package_name}/{service_name}')

    for package_name, message_names in get_message_interfaces().items():
        for message_name in message_names:
            types.append(f'{package_name}/{message_name}')

    for package_name, action_names in get_action_interfaces().items():
        for action_name in action_names:
            types.append(f'{package_name}/{action_name}')
    return sorted(types)


def interface_to_yaml(identifier):
    interface = utilities.get_interface(identifier)
    if utilities.is_action(interface):
        instance = interface.Goal()
    elif utilities.is_service(interface):
        instance = interface.Request()
    else:
        instance = interface()

    return message_to_yaml(instance)
